from qtpy.QtCore import QRegExp
from qtpy.QtGui import QRegExpValidator
from qtpy.QtWidgets import QDialog

from .redis import QRedis
from .qutil import ui_loadable


@ui_loadable
class OpenRedisDialog(QDialog):

    Dialog = None

    def __init__(self, parent=None):
        """Display a dialog to open a new redis connection"""
        super(OpenRedisDialog, self).__init__(parent)
        self.load_ui()
        v = QRegExpValidator(QRegExp("([^:/]+)?(:[0-9]+)?"))
        self.ui.tcp.setValidator(v)
        self.ui.tcp_option.setChecked(True)

    def _create_redis(self):
        if self.exec_() != QDialog.Accepted:
            return None, None
        kwargs = dict(db=self.ui.db.value())
        client_name = self.ui.name.text()
        user = self.ui.user.text()
        password = self.ui.password.text()
        if client_name:
            kwargs["client_name"] = client_name
        if user:
            kwargs["username"] = user
        if password:
            kwargs["password"] = password
        if self.ui.tcp_option.isChecked():
            url = self.ui.tcp.text()
            if ":" in url:
                host, port = url.split(":")
                if port:
                    kwargs["port"] = int(port)
            else:
                host = url
            if host:
                kwargs["host"] = host
        else:
            kwargs["unix_socket_path"] = self.ui.socket.text()
        opts = dict(
            filter=self.ui.filter.text(),
            split_by=self.ui.splitter.text(),
        )
        return QRedis(**kwargs), opts

    @classmethod
    def create_redis(cls, parent=None):
        if cls.Dialog is None:
            cls.Dialog = cls(parent=parent)
        return cls.Dialog._create_redis()


class AboutDialog(QDialog):
    """Create the necessary elements to show helpful text in a dialog."""

    def __init__(self, parent=None):
        """Display a dialog that shows application information."""
        super(AboutDialog, self).__init__(parent)
        self.setWindowTitle("About")



import pickle
import collections

import msgpack
import msgpack_numpy
from redis import Redis
from qtpy.QtCore import QObject, Signal

from .util import KeyItem


def msgpack_pack(data):
    return msgpack.packb(data, use_bin_type=True, default=msgpack_numpy.encode)


def msgpack_unpack(buff):
    return msgpack.unpackb(buff, raw=False, object_hook=msgpack_numpy.decode)


def decode_utf8(v):
    return v.decode()


def decode_pickle(v):
    return str(pickle.loads(v))


def decode_msgpack(v):
    return str(msgpack_unpack(v))


DECODES = [
    (decode_utf8, "utf-8"),
    (decode_pickle, "pickle"),
    (decode_msgpack, "msgpack"),
    (str, "raw"),
]


def decode(value):
    for decoder, dtype in DECODES:
        try:
            return decoder(value)
        except Exception:
            continue


def _set(redis, key, value):
    redis.set(key, value)


def _set_list(redis, key, lst):
    redis.delete(key)
    redis.rpush(key, *lst)


def _set_set(redis, key, st):
    redis.delete(key)
    redis.sadd(key, *tuple(st))


class zset(list):
    pass


class stream(tuple):
    pass


class QRedis(QObject):

    keyRenamed = Signal(object, object)
    keysDeleted = Signal()

    TYPE_MAP = {
        type(None): "none",
        str: "string",
        dict: "hash",
        list: "list",
        set: "set",
        zset: "zset",
        stream: "stream"
    }

    def __init__(self, *args, **kwargs):
        if args:
            parent, args = args[0], args[1:]
        else:
            parent = kwargs.pop("parent", None)
        # kwargs.setdefault("decode_responses", True)
        super(QRedis, self).__init__(parent)

        self._get_type_map = {
            "none": lambda v: None,
            "string": self._get,
            "hash": self._hgetall,
            "list": self._lgetall,
            "set": self._sgetall,
            "zset": self._zgetall,
            "stream": self._xrange,
        }

        self._set_type_map = collections.defaultdict(
            lambda: lambda r, k, v: _set(r.redis, k, v),
            {
                type(None): lambda k, v: self.delete(k),
                dict: lambda k, v: self.redis.hmset(k, v),
                list: lambda k, v: _set_list(self.redis, k, v),
                set: lambda k, v: _set_set(self.redis, k, v),
            },
        )

        self.redis = Redis(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.redis, name)

    def __getitem__(self, key):
        value = self.get(key)
        if value is None:
            raise KeyError(key)
        return value

    def __setitem__(self, key, value):

        self._set_type_map[type(value)](self, key, value)

    def __delitem__(self, key):
        self.delete(key)

    def _get(self, key):
        return decode(self.redis.get(key))

    def _hgetall(self, key):
        return {decode(k): decode(v) for k, v in self.redis.hgetall(key).items()}

    def _lgetall(self, key):
        return [decode(i) for i in self.redis.lrange(key, 0, -1)]

    def _sgetall(self, key):
        return {decode(i) for i in self.redis.smembers(key)}

    def _zgetall(self, key):
        return {
            decode(member): decode(score)
            for member, score in self.redis.zscan_iter(key)
        }

    def _xrange(self, key):
        data = []
        for i in self.redis.xrange(key):
            event_time, event_data_raw = i
            event_time = decode(event_time)
            event_data = {
                decode(member): decode(score)
                for member, score in event_data_raw.items()
            }
            data.append((event_time, event_data))
        return data
    
    def get(self, key, default=None):
        if not self.exists(key):
            return default
        dtype, ttl = self.type(key), self.ttl(key)
        ttl = -1 if ttl is None else ttl  # handle redis < 2.8
        value = self._get_type_map[dtype](key)
        return KeyItem(self, key, dtype, ttl, value)

    def type(self, name):
        return self.redis.type(name).decode()

    def keys(self, pattern):
        return [k.decode() for k in self.redis.keys(pattern)]

    def has_key(self, key):
        return self.exists(key)

    def delete(self, *keys):
        self.redis.delete(*keys)
        self.keysDeleted.emit()

    def rename(self, old_key, new_key):
        old_item = self[old_key]
        self.redis.rename(old_key, new_key)
        new_item = self[new_key]
        self.keyRenamed.emit(old_item, new_item)



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QSplitter

from .tree import RedisTree
from .editor import RedisEditor


class RedisPanel(QSplitter):
    def __init__(self, redis, parent=None):
        super(RedisPanel, self).__init__(parent)
        self.redis = redis
        self.tree = RedisTree(redis, parent=self)
        self.editor = RedisEditor(self)
        self.tree.setWindowFlags(Qt.Widget)
        self.addWidget(self.tree)
        self.addWidget(self.editor)
        self.setStretchFactor(0, 1)
        self.setStretchFactor(1, 1)
        self.tree.currentChanged.connect(self.__on_selection_changed)
        self.tree.addKey.connect(self.editor.set_item)

    def __on_add_key(self, item):
        self.editor.set_item(item)

    def __on_selection_changed(self, node):
        if node is None:
            self.editor.set_empty()
        elif node.is_key():
            item = self.redis.get(node.key)
            self.editor.set_item(item)
        elif node.is_db():
            self.editor.set_db(self.redis)
